from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

from sqlalchemy import bindparam, delete, func, insert, select, update
from sqlalchemy.exc import IntegrityError, NoResultFound

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import BindParameter

from diracx.core.exceptions import InvalidQueryError, JobNotFound
from diracx.core.models import (
    JobMinorStatus,
    JobStatus,
    LimitedJobStatusReturn,
    ScalarSearchOperator,
    ScalarSearchSpec,
    SearchSpec,
    SortSpec,
)

from ..utils import BaseSQLDB, apply_search_filters, apply_sort_constraints
from .schema import (
    InputData,
    JobCommands,
    JobDBBase,
    JobJDLs,
    Jobs,
)


def _get_columns(table, parameters):
    columns = [x for x in table.columns]
    if parameters:
        if unrecognised_parameters := set(parameters) - set(table.columns.keys()):
            raise InvalidQueryError(
                f"Unrecognised parameters requested {unrecognised_parameters}"
            )
        columns = [c for c in columns if c.name in parameters]
    return columns


class JobDB(BaseSQLDB):
    metadata = JobDBBase.metadata

    # TODO: this is copied from the DIRAC JobDB
    # but is overwriten in LHCbDIRAC, so we need
    # to find a way to make it dynamic
    jdl2DBParameters = ["JobName", "JobType", "JobGroup"]

    # TODO: set maxRescheduling value from CS
    # maxRescheduling = self.getCSOption("MaxRescheduling", 3)
    # For now:
    maxRescheduling = 3

    async def summary(self, group_by, search) -> list[dict[str, str | int]]:
        columns = _get_columns(Jobs.__table__, group_by)

        stmt = select(*columns, func.count(Jobs.JobID).label("count"))
        stmt = apply_search_filters(Jobs.__table__.columns.__getitem__, stmt, search)
        stmt = stmt.group_by(*columns)

        # Execute the query
        return [
            dict(row._mapping)
            async for row in (await self.conn.stream(stmt))
            if row.count > 0  # type: ignore
        ]

    async def search(
        self,
        parameters: list[str] | None,
        search: list[SearchSpec],
        sorts: list[SortSpec],
        *,
        distinct: bool = False,
        per_page: int = 100,
        page: int | None = None,
    ) -> tuple[int, list[dict[Any, Any]]]:
        # Find which columns to select
        columns = _get_columns(Jobs.__table__, parameters)
        stmt = select(*columns)

        stmt = apply_search_filters(Jobs.__table__.columns.__getitem__, stmt, search)
        stmt = apply_sort_constraints(Jobs.__table__.columns.__getitem__, stmt, sorts)

        if distinct:
            stmt = stmt.distinct()

        # Calculate total count before applying pagination
        total_count_subquery = stmt.alias()
        total_count_stmt = select(func.count()).select_from(total_count_subquery)
        total = (await self.conn.execute(total_count_stmt)).scalar_one()

        # Apply pagination
        if page is not None:
            if page < 1:
                raise InvalidQueryError("Page must be a positive integer")
            if per_page < 1:
                raise InvalidQueryError("Per page must be a positive integer")
            stmt = stmt.offset((page - 1) * per_page).limit(per_page)

        # Execute the query
        return total, [
            dict(row._mapping) async for row in (await self.conn.stream(stmt))
        ]

    async def _insertNewJDL(self, jdl) -> int:
        from DIRAC.WorkloadManagementSystem.DB.JobDBUtils import compressJDL

        stmt = insert(JobJDLs).values(
            JDL="", JobRequirements="", OriginalJDL=compressJDL(jdl)
        )
        result = await self.conn.execute(stmt)
        # await self.engine.commit()
        return result.lastrowid

    async def _insertJob(self, jobData: dict[str, Any]):
        stmt = insert(Jobs).values(jobData)
        await self.conn.execute(stmt)

    async def _insertInputData(self, job_id: int, lfns: list[str]):
        stmt = insert(InputData).values([{"JobID": job_id, "LFN": lfn} for lfn in lfns])
        await self.conn.execute(stmt)

    async def setJobAttributes(self, job_id, jobData):
        """TODO: add myDate and force parameters."""
        if "Status" in jobData:
            jobData = jobData | {"LastUpdateTime": datetime.now(tz=timezone.utc)}
        stmt = update(Jobs).where(Jobs.JobID == job_id).values(jobData)
        await self.conn.execute(stmt)

    async def _checkAndPrepareJob(
        self,
        jobID,
        class_ad_job,
        class_ad_req,
        owner,
        owner_group,
        job_attrs,
        vo,
    ):
        """Check Consistency of Submitted JDL and set some defaults
        Prepare subJDL with Job Requirements.
        """
        from DIRAC.Core.Utilities.DErrno import EWMSSUBM, cmpError
        from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
        from DIRAC.WorkloadManagementSystem.DB.JobDBUtils import (
            checkAndPrepareJob,
        )

        retVal = checkAndPrepareJob(
            jobID,
            class_ad_job,
            class_ad_req,
            owner,
            owner_group,
            job_attrs,
            vo,
        )

        if not retVal["OK"]:
            if cmpError(retVal, EWMSSUBM):
                await self.setJobAttributes(jobID, job_attrs)

            returnValueOrRaise(retVal)

    async def setJobJDL(self, job_id, jdl):
        from DIRAC.WorkloadManagementSystem.DB.JobDBUtils import compressJDL

        stmt = (
            update(JobJDLs).where(JobJDLs.JobID == job_id).values(JDL=compressJDL(jdl))
        )
        await self.conn.execute(stmt)

    async def getJobJDL(self, job_id: int, original: bool = False) -> str:
        from DIRAC.WorkloadManagementSystem.DB.JobDBUtils import extractJDL

        if original:
            stmt = select(JobJDLs.OriginalJDL).where(JobJDLs.JobID == job_id)
        else:
            stmt = select(JobJDLs.JDL).where(JobJDLs.JobID == job_id)

        jdl = (await self.conn.execute(stmt)).scalar_one()
        if jdl:
            jdl = extractJDL(jdl)

        return jdl

    async def insert(
        self,
        jdl,
        owner,
        owner_group,
        initial_status,
        initial_minor_status,
        vo,
    ):
        from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
        from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
        from DIRAC.WorkloadManagementSystem.DB.JobDBUtils import (
            checkAndAddOwner,
            createJDLWithInitialStatus,
            fixJDL,
        )

        job_attrs = {
            "LastUpdateTime": datetime.now(tz=timezone.utc),
            "SubmissionTime": datetime.now(tz=timezone.utc),
            "Owner": owner,
            "OwnerGroup": owner_group,
            "VO": vo,
        }

        jobManifest = returnValueOrRaise(checkAndAddOwner(jdl, owner, owner_group))

        jdl = fixJDL(jdl)

        job_id = await self._insertNewJDL(jdl)

        jobManifest.setOption("JobID", job_id)

        job_attrs["JobID"] = job_id

        # 2.- Check JDL and Prepare DIRAC JDL
        jobJDL = jobManifest.dumpAsJDL()

        # Replace the JobID placeholder if any
        if jobJDL.find("%j") != -1:
            jobJDL = jobJDL.replace("%j", str(job_id))

        class_ad_job = ClassAd(jobJDL)
        class_ad_req = ClassAd("[]")
        if not class_ad_job.isOK():
            job_attrs["Status"] = JobStatus.FAILED

            job_attrs["MinorStatus"] = "Error in JDL syntax"

            await self._insertJob(job_attrs)

            return {
                "JobID": job_id,
                "Status": JobStatus.FAILED,
                "MinorStatus": "Error in JDL syntax",
            }

        class_ad_job.insertAttributeInt("JobID", job_id)

        await self._checkAndPrepareJob(
            job_id,
            class_ad_job,
            class_ad_req,
            owner,
            owner_group,
            job_attrs,
            vo,
        )

        jobJDL = createJDLWithInitialStatus(
            class_ad_job,
            class_ad_req,
            self.jdl2DBParameters,
            job_attrs,
            initial_status,
            initial_minor_status,
            modern=True,
        )

        await self.setJobJDL(job_id, jobJDL)

        # Adding the job in the Jobs table
        await self._insertJob(job_attrs)

        # TODO: check if that is actually true
        if class_ad_job.lookupAttribute("Parameters"):
            raise NotImplementedError("Parameters in the JDL are not supported")

        # Looking for the Input Data
        inputData = []
        if class_ad_job.lookupAttribute("InputData"):
            inputData = class_ad_job.getListFromExpression("InputData")
            lfns = [lfn for lfn in inputData if lfn]
            if lfns:
                await self._insertInputData(job_id, lfns)

        return {
            "JobID": job_id,
            "Status": initial_status,
            "MinorStatus": initial_minor_status,
            "TimeStamp": datetime.now(tz=timezone.utc),
        }

    async def rescheduleJob(self, job_id) -> dict[str, Any]:
        """Reschedule given job."""
        from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
        from DIRAC.Core.Utilities.ReturnValues import SErrorException

        _, result = await self.search(
            parameters=[
                "Status",
                "MinorStatus",
                "VerifiedFlag",
                "RescheduleCounter",
                "Owner",
                "OwnerGroup",
            ],
            search=[
                ScalarSearchSpec(
                    parameter="JobID", operator=ScalarSearchOperator.EQUAL, value=job_id
                )
            ],
            sorts=[],
        )
        if not result:
            raise ValueError(f"Job {job_id} not found.")

        jobAttrs = result[0]

        if "VerifiedFlag" not in jobAttrs:
            raise ValueError(f"Job {job_id} not found in the system")

        if not jobAttrs["VerifiedFlag"]:
            raise ValueError(
                f"Job {job_id} not Verified: Status {jobAttrs['Status']}, Minor Status: {jobAttrs['MinorStatus']}"
            )

        reschedule_counter = int(jobAttrs["RescheduleCounter"]) + 1

        # TODO: update maxRescheduling:
        # self.maxRescheduling = self.getCSOption("MaxRescheduling", self.maxRescheduling)

        if reschedule_counter > self.maxRescheduling:
            logging.warn(f"Job {job_id}: Maximum number of reschedulings is reached.")
            self.setJobAttributes(
                job_id,
                {
                    "Status": JobStatus.FAILED,
                    "MinorStatus": JobMinorStatus.MAX_RESCHEDULING,
                },
            )
            raise ValueError(
                f"Maximum number of reschedulings is reached: {self.maxRescheduling}"
            )

        new_job_attributes = {"RescheduleCounter": reschedule_counter}

        # TODO: get the job parameters from JobMonitoringClient
        # result = JobMonitoringClient().getJobParameters(jobID)
        # if result["OK"]:
        #     parDict = result["Value"]
        #     for key, value in parDict.get(jobID, {}).items():
        #         result = self.setAtticJobParameter(jobID, key, value, rescheduleCounter - 1)
        #         if not result["OK"]:
        #             break

        # TODO: IF we keep JobParameters and OptimizerParameters: Delete job in those tables.
        # await self.delete_job_parameters(job_id)
        # await self.delete_job_optimizer_parameters(job_id)

        job_jdl = await self.getJobJDL(job_id, original=True)
        if not job_jdl.strip().startswith("["):
            job_jdl = f"[{job_jdl}]"

        classAdJob = ClassAd(job_jdl)
        classAdReq = ClassAd("[]")
        retVal = {}
        retVal["JobID"] = job_id

        classAdJob.insertAttributeInt("JobID", job_id)

        try:
            result = await self._checkAndPrepareJob(
                job_id,
                classAdJob,
                classAdReq,
                jobAttrs["Owner"],
                jobAttrs["OwnerGroup"],
                new_job_attributes,
                classAdJob.getAttributeString("VirtualOrganization"),
            )
        except SErrorException as e:
            raise ValueError(e) from e

        priority = classAdJob.getAttributeInt("Priority")
        if priority is None:
            priority = 0
        jobAttrs["UserPriority"] = priority

        siteList = classAdJob.getListFromExpression("Site")
        if not siteList:
            site = "ANY"
        elif len(siteList) > 1:
            site = "Multiple"
        else:
            site = siteList[0]

        jobAttrs["Site"] = site

        jobAttrs["Status"] = JobStatus.RECEIVED

        jobAttrs["MinorStatus"] = JobMinorStatus.RESCHEDULED

        jobAttrs["ApplicationStatus"] = "Unknown"

        jobAttrs["LastUpdateTime"] = datetime.now(tz=timezone.utc)

        jobAttrs["RescheduleTime"] = datetime.now(tz=timezone.utc)

        reqJDL = classAdReq.asJDL()
        classAdJob.insertAttributeInt("JobRequirements", reqJDL)

        jobJDL = classAdJob.asJDL()

        # Replace the JobID placeholder if any
        jobJDL = jobJDL.replace("%j", str(job_id))

        result = await self.setJobJDL(job_id, jobJDL)

        result = await self.setJobAttributes(job_id, jobAttrs)

        retVal["InputData"] = classAdJob.lookupAttribute("InputData")
        retVal["RescheduleCounter"] = reschedule_counter
        retVal["Status"] = JobStatus.RECEIVED
        retVal["MinorStatus"] = JobMinorStatus.RESCHEDULED

        return retVal

    async def get_job_status(self, job_id: int) -> LimitedJobStatusReturn:
        try:
            stmt = select(Jobs.Status, Jobs.MinorStatus, Jobs.ApplicationStatus).where(
                Jobs.JobID == job_id
            )
            return LimitedJobStatusReturn(
                **dict((await self.conn.execute(stmt)).one()._mapping)
            )
        except NoResultFound as e:
            raise JobNotFound(job_id) from e

    async def set_job_command(self, job_id: int, command: str, arguments: str = ""):
        """Store a command to be passed to the job together with the next heart beat."""
        try:
            stmt = insert(JobCommands).values(
                JobID=job_id,
                Command=command,
                Arguments=arguments,
                ReceptionTime=datetime.now(tz=timezone.utc),
            )
            await self.conn.execute(stmt)
        except IntegrityError as e:
            raise JobNotFound(job_id) from e

    async def delete_jobs(self, job_ids: list[int]):
        """Delete jobs from the database."""
        stmt = delete(JobJDLs).where(JobJDLs.JobID.in_(job_ids))
        await self.conn.execute(stmt)

    async def set_properties(
        self, properties: dict[int, dict[str, Any]], update_timestamp: bool = False
    ) -> int:
        """Update the job parameters
        All the jobs must update the same properties.

        :param properties: {job_id : {prop1: val1, prop2:val2}
        :param update_timestamp: if True, update the LastUpdate to now

        :return rowcount

        """
        # Check that all we always update the same set of properties
        required_parameters_set = {tuple(sorted(k.keys())) for k in properties.values()}

        if len(required_parameters_set) != 1:
            raise NotImplementedError(
                "All the jobs should update the same set of properties"
            )

        required_parameters = list(required_parameters_set)[0]
        update_parameters = [{"job_id": k, **v} for k, v in properties.items()]

        columns = _get_columns(Jobs.__table__, required_parameters)
        values: dict[str, BindParameter[Any] | datetime] = {
            c.name: bindparam(c.name) for c in columns
        }
        if update_timestamp:
            values["LastUpdateTime"] = datetime.now(tz=timezone.utc)

        stmt = update(Jobs).where(Jobs.JobID == bindparam("job_id")).values(**values)
        rows = await self.conn.execute(stmt, update_parameters)

        return rows.rowcount



from sqlalchemy import (
    DateTime,
    Double,
    Index,
    Integer,
    String,
    Text,
)
from sqlalchemy.orm import declarative_base

from ..utils import Column, EnumBackedBool, NullColumn

PilotAgentsDBBase = declarative_base()


class PilotAgents(PilotAgentsDBBase):
    __tablename__ = "PilotAgents"

    PilotID = Column("PilotID", Integer, autoincrement=True, primary_key=True)
    InitialJobID = Column("InitialJobID", Integer, default=0)
    CurrentJobID = Column("CurrentJobID", Integer, default=0)
    PilotJobReference = Column("PilotJobReference", String(255), default="Unknown")
    PilotStamp = Column("PilotStamp", String(32), default="")
    DestinationSite = Column("DestinationSite", String(128), default="NotAssigned")
    Queue = Column("Queue", String(128), default="Unknown")
    GridSite = Column("GridSite", String(128), default="Unknown")
    VO = Column("VO", String(128))
    GridType = Column("GridType", String(32), default="LCG")
    BenchMark = Column("BenchMark", Double, default=0.0)
    SubmissionTime = NullColumn("SubmissionTime", DateTime)
    LastUpdateTime = NullColumn("LastUpdateTime", DateTime)
    Status = Column("Status", String(32), default="Unknown")
    StatusReason = Column("StatusReason", String(255), default="Unknown")
    AccountingSent = Column("AccountingSent", EnumBackedBool(), default=False)

    __table_args__ = (
        Index("PilotJobReference", "PilotJobReference"),
        Index("Status", "Status"),
        Index("Statuskey", "GridSite", "DestinationSite", "Status"),
    )


class JobToPilotMapping(PilotAgentsDBBase):
    __tablename__ = "JobToPilotMapping"

    PilotID = Column("PilotID", Integer, primary_key=True)
    JobID = Column("JobID", Integer, primary_key=True)
    StartTime = Column("StartTime", DateTime)

    __table_args__ = (Index("JobID", "JobID"), Index("PilotID", "PilotID"))


class PilotOutput(PilotAgentsDBBase):
    __tablename__ = "PilotOutput"

    PilotID = Column("PilotID", Integer, primary_key=True)
    StdOutput = Column("StdOutput", Text)
    StdError = Column("StdError", Text)



from sqlalchemy import (
    Integer,
    Numeric,
    PrimaryKeyConstraint,
    String,
)
from sqlalchemy.orm import declarative_base

from ..utils import Column, DateNowColumn

JobLoggingDBBase = declarative_base()


class LoggingInfo(JobLoggingDBBase):
    __tablename__ = "LoggingInfo"
    JobID = Column(Integer)
    SeqNum = Column(Integer)
    Status = Column(String(32), default="")
    MinorStatus = Column(String(128), default="")
    ApplicationStatus = Column(String(255), default="")
    StatusTime = DateNowColumn()
    # TODO: Check that this corresponds to the DOUBLE(12,3) type in MySQL
    StatusTimeOrder = Column(Numeric(precision=12, scale=3), default=0)
    Source = Column(String(32), default="Unknown", name="StatusSource")
    __table_args__ = (PrimaryKeyConstraint("JobID", "SeqNum"),)



from __future__ import annotations

import asyncio

import pytest

from diracx.core.exceptions import InvalidQueryError, JobNotFound
from diracx.core.models import (
    ScalarSearchOperator,
    ScalarSearchSpec,
    SortDirection,
    SortSpec,
    VectorSearchOperator,
    VectorSearchSpec,
)
from diracx.db.sql.job.db import JobDB


@pytest.fixture
async def job_db(tmp_path):
    job_db = JobDB("sqlite+aiosqlite:///:memory:")
    async with job_db.engine_context():
        async with job_db.engine.begin() as conn:
            # set PRAGMA foreign_keys=ON if sqlite
            if job_db._db_url.startswith("sqlite"):
                await conn.exec_driver_sql("PRAGMA foreign_keys=ON")
            await conn.run_sync(job_db.metadata.create_all)
        yield job_db


async def test_search_parameters(job_db):
    """Test that we can search specific parameters for jobs in the database."""
    async with job_db as job_db:
        total, result = await job_db.search(["JobID"], [], [])
        assert total == 0
        assert not result

        result = await asyncio.gather(
            *(
                job_db.insert(
                    f"JDL{i}",
                    "owner",
                    "owner_group",
                    "New",
                    "dfdfds",
                    "lhcb",
                )
                for i in range(100)
            )
        )

    async with job_db as job_db:
        # Search a specific parameter: JobID
        total, result = await job_db.search(["JobID"], [], [])
        assert total == 100
        assert result
        for r in result:
            assert r.keys() == {"JobID"}

        # Search a specific parameter: Status
        total, result = await job_db.search(["Status"], [], [])
        assert total == 100
        assert result
        for r in result:
            assert r.keys() == {"Status"}

        # Search for multiple parameters: JobID, Status
        total, result = await job_db.search(["JobID", "Status"], [], [])
        assert total == 100
        assert result
        for r in result:
            assert r.keys() == {"JobID", "Status"}

        # Search for a specific parameter but use distinct: Status
        total, result = await job_db.search(["Status"], [], [], distinct=True)
        assert total == 1
        assert result

        # Search for a non-existent parameter: Dummy
        with pytest.raises(InvalidQueryError):
            total, result = await job_db.search(["Dummy"], [], [])


async def test_search_conditions(job_db):
    """Test that we can search for specific jobs in the database."""
    async with job_db as job_db:
        result = await asyncio.gather(
            *(
                job_db.insert(
                    f"JDL{i}",
                    f"owner{i}",
                    "owner_group",
                    "New",
                    "dfdfds",
                    "lhcb",
                )
                for i in range(100)
            )
        )

    async with job_db as job_db:
        # Search a specific scalar condition: JobID eq 3
        condition = ScalarSearchSpec(
            parameter="JobID", operator=ScalarSearchOperator.EQUAL, value=3
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 1
        assert result
        assert len(result) == 1
        assert result[0]["JobID"] == 3

        # Search a specific scalar condition: JobID lt 3
        condition = ScalarSearchSpec(
            parameter="JobID", operator=ScalarSearchOperator.LESS_THAN, value=3
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 2
        assert result
        assert len(result) == 2
        assert result[0]["JobID"] == 1
        assert result[1]["JobID"] == 2

        # Search a specific scalar condition: JobID neq 3
        condition = ScalarSearchSpec(
            parameter="JobID", operator=ScalarSearchOperator.NOT_EQUAL, value=3
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 99
        assert result
        assert len(result) == 99
        assert all(r["JobID"] != 3 for r in result)

        # Search a specific scalar condition: JobID eq 5873 (does not exist)
        condition = ScalarSearchSpec(
            parameter="JobID", operator=ScalarSearchOperator.EQUAL, value=5873
        )
        total, result = await job_db.search([], [condition], [])
        assert not result

        # Search a specific vector condition: JobID in 1,2,3
        condition = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.IN, values=[1, 2, 3]
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 3
        assert result
        assert len(result) == 3
        assert all(r["JobID"] in [1, 2, 3] for r in result)

        # Search a specific vector condition: JobID in 1,2,5873 (one of them does not exist)
        condition = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.IN, values=[1, 2, 5873]
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 2
        assert result
        assert len(result) == 2
        assert all(r["JobID"] in [1, 2] for r in result)

        # Search a specific vector condition: JobID not in 1,2,3
        condition = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.NOT_IN, values=[1, 2, 3]
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 97
        assert result
        assert len(result) == 97
        assert all(r["JobID"] not in [1, 2, 3] for r in result)

        # Search a specific vector condition: JobID not in 1,2,5873 (one of them does not exist)
        condition = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.NOT_IN, values=[1, 2, 5873]
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 98
        assert result
        assert len(result) == 98
        assert all(r["JobID"] not in [1, 2] for r in result)

        # Search for multiple conditions based on different parameters: JobID eq 70, JobID in 4,5,6
        condition1 = ScalarSearchSpec(
            parameter="Owner", operator=ScalarSearchOperator.EQUAL, value="owner4"
        )
        condition2 = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.IN, values=[4, 5, 6]
        )
        total, result = await job_db.search([], [condition1, condition2], [])
        assert total == 1
        assert result
        assert len(result) == 1
        assert result[0]["JobID"] == 5
        assert result[0]["Owner"] == "owner4"

        # Search for multiple conditions based on the same parameter: JobID eq 70, JobID in 4,5,6
        condition1 = ScalarSearchSpec(
            parameter="JobID", operator=ScalarSearchOperator.EQUAL, value=70
        )
        condition2 = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.IN, values=[4, 5, 6]
        )
        total, result = await job_db.search([], [condition1, condition2], [])
        assert total == 0
        assert not result


async def test_search_sorts(job_db):
    """Test that we can search for jobs in the database and sort the results."""
    async with job_db as job_db:
        result = await asyncio.gather(
            *(
                job_db.insert(
                    f"JDL{i}",
                    f"owner{i}",
                    "owner_group1" if i < 50 else "owner_group2",
                    "New",
                    "dfdfds",
                    "lhcb",
                )
                for i in range(100)
            )
        )

    async with job_db as job_db:
        # Search and sort by JobID in ascending order
        sort = SortSpec(parameter="JobID", direction=SortDirection.ASC)
        total, result = await job_db.search([], [], [sort])
        assert total == 100
        assert result
        for i, r in enumerate(result):
            assert r["JobID"] == i + 1

        # Search and sort by JobID in descending order
        sort = SortSpec(parameter="JobID", direction=SortDirection.DESC)
        total, result = await job_db.search([], [], [sort])
        assert total == 100
        assert result
        for i, r in enumerate(result):
            assert r["JobID"] == 100 - i

        # Search and sort by Owner in ascending order
        sort = SortSpec(parameter="Owner", direction=SortDirection.ASC)
        total, result = await job_db.search([], [], [sort])
        assert total == 100
        assert result
        # Assert that owner10 is before owner2 because of the lexicographical order
        assert result[2]["Owner"] == "owner10"
        assert result[12]["Owner"] == "owner2"

        # Search and sort by Owner in descending order
        sort = SortSpec(parameter="Owner", direction=SortDirection.DESC)
        total, result = await job_db.search([], [], [sort])
        assert total == 100
        assert result
        # Assert that owner10 is before owner2 because of the lexicographical order
        assert result[97]["Owner"] == "owner10"
        assert result[87]["Owner"] == "owner2"

        # Search and sort by OwnerGroup in ascending order and JobID in descending order
        sort1 = SortSpec(parameter="OwnerGroup", direction=SortDirection.ASC)
        sort2 = SortSpec(parameter="JobID", direction=SortDirection.DESC)
        total, result = await job_db.search([], [], [sort1, sort2])
        assert total == 100
        assert result
        assert result[0]["OwnerGroup"] == "owner_group1"
        assert result[0]["JobID"] == 50
        assert result[99]["OwnerGroup"] == "owner_group2"
        assert result[99]["JobID"] == 51


async def test_search_pagination(job_db):
    """Test that we can search for jobs in the database."""
    async with job_db as job_db:
        result = await asyncio.gather(
            *(
                job_db.insert(
                    f"JDL{i}",
                    f"owner{i}",
                    "owner_group1" if i < 50 else "owner_group2",
                    "New",
                    "dfdfds",
                    "lhcb",
                )
                for i in range(100)
            )
        )

    async with job_db as job_db:
        # Search for the first 10 jobs
        total, result = await job_db.search([], [], [], per_page=10, page=1)
        assert total == 100
        assert result
        assert len(result) == 10
        assert result[0]["JobID"] == 1

        # Search for the second 10 jobs
        total, result = await job_db.search([], [], [], per_page=10, page=2)
        assert total == 100
        assert result
        assert len(result) == 10
        assert result[0]["JobID"] == 11

        # Search for the last 10 jobs
        total, result = await job_db.search([], [], [], per_page=10, page=10)
        assert total == 100
        assert result
        assert len(result) == 10
        assert result[0]["JobID"] == 91

        # Search for the second 50 jobs
        total, result = await job_db.search([], [], [], per_page=50, page=2)
        assert total == 100
        assert result
        assert len(result) == 50
        assert result[0]["JobID"] == 51

        # Invalid page number
        total, result = await job_db.search([], [], [], per_page=10, page=11)
        assert total == 100
        assert not result

        # Invalid page number
        with pytest.raises(InvalidQueryError):
            result = await job_db.search([], [], [], per_page=10, page=0)

        # Invalid per_page number
        with pytest.raises(InvalidQueryError):
            result = await job_db.search([], [], [], per_page=0, page=1)


async def test_set_job_command_invalid_job_id(job_db: JobDB):
    """Test that setting a command for a non-existent job raises JobNotFound."""
    async with job_db as job_db:
        with pytest.raises(JobNotFound):
            await job_db.set_job_command(123456, "test_command")



from __future__ import annotations

from typing import Any

import sqlalchemy

from diracx.core.models import SandboxInfo, SandboxType, UserInfo
from diracx.db.sql.utils import BaseSQLDB, utcnow

from .schema import Base as SandboxMetadataDBBase
from .schema import sb_EntityMapping, sb_Owners, sb_SandBoxes


class SandboxMetadataDB(BaseSQLDB):
    metadata = SandboxMetadataDBBase.metadata

    async def upsert_owner(self, user: UserInfo) -> int:
        """Get the id of the owner from the database."""
        # TODO: Follow https://github.com/DIRACGrid/diracx/issues/49
        stmt = sqlalchemy.select(sb_Owners.OwnerID).where(
            sb_Owners.Owner == user.preferred_username,
            sb_Owners.OwnerGroup == user.dirac_group,
            sb_Owners.VO == user.vo,
        )
        result = await self.conn.execute(stmt)
        if owner_id := result.scalar_one_or_none():
            return owner_id

        stmt = sqlalchemy.insert(sb_Owners).values(
            Owner=user.preferred_username,
            OwnerGroup=user.dirac_group,
            VO=user.vo,
        )
        result = await self.conn.execute(stmt)
        return result.lastrowid

    @staticmethod
    def get_pfn(bucket_name: str, user: UserInfo, sandbox_info: SandboxInfo) -> str:
        """Get the sandbox's user namespaced and content addressed PFN."""
        parts = [
            "S3",
            bucket_name,
            user.vo,
            user.dirac_group,
            user.preferred_username,
            f"{sandbox_info.checksum_algorithm}:{sandbox_info.checksum}.{sandbox_info.format}",
        ]
        return "/" + "/".join(parts)

    async def insert_sandbox(
        self, se_name: str, user: UserInfo, pfn: str, size: int
    ) -> None:
        """Add a new sandbox in SandboxMetadataDB."""
        # TODO: Follow https://github.com/DIRACGrid/diracx/issues/49
        owner_id = await self.upsert_owner(user)
        stmt = sqlalchemy.insert(sb_SandBoxes).values(
            OwnerId=owner_id,
            SEName=se_name,
            SEPFN=pfn,
            Bytes=size,
            RegistrationTime=utcnow(),
            LastAccessTime=utcnow(),
        )
        try:
            result = await self.conn.execute(stmt)
        except sqlalchemy.exc.IntegrityError:
            await self.update_sandbox_last_access_time(se_name, pfn)
        else:
            assert result.rowcount == 1

    async def update_sandbox_last_access_time(self, se_name: str, pfn: str) -> None:
        stmt = (
            sqlalchemy.update(sb_SandBoxes)
            .where(sb_SandBoxes.SEName == se_name, sb_SandBoxes.SEPFN == pfn)
            .values(LastAccessTime=utcnow())
        )
        result = await self.conn.execute(stmt)
        assert result.rowcount == 1

    async def sandbox_is_assigned(self, pfn: str, se_name: str) -> bool:
        """Checks if a sandbox exists and has been assigned."""
        stmt: sqlalchemy.Executable = sqlalchemy.select(sb_SandBoxes.Assigned).where(
            sb_SandBoxes.SEName == se_name, sb_SandBoxes.SEPFN == pfn
        )
        result = await self.conn.execute(stmt)
        is_assigned = result.scalar_one()
        return is_assigned

    @staticmethod
    def jobid_to_entity_id(job_id: int) -> str:
        """Define the entity id as 'Entity:entity_id' due to the DB definition."""
        return f"Job:{job_id}"

    async def get_sandbox_assigned_to_job(
        self, job_id: int, sb_type: SandboxType
    ) -> list[Any]:
        """Get the sandbox assign to job."""
        entity_id = self.jobid_to_entity_id(job_id)
        stmt = (
            sqlalchemy.select(sb_SandBoxes.SEPFN)
            .where(sb_SandBoxes.SBId == sb_EntityMapping.SBId)
            .where(
                sb_EntityMapping.EntityId == entity_id,
                sb_EntityMapping.Type == sb_type,
            )
        )
        result = await self.conn.execute(stmt)
        return [result.scalar()]

    async def assign_sandbox_to_jobs(
        self,
        jobs_ids: list[int],
        pfn: str,
        sb_type: SandboxType,
        se_name: str,
    ) -> None:
        """Mapp sandbox and jobs."""
        for job_id in jobs_ids:
            # Define the entity id as 'Entity:entity_id' due to the DB definition:
            entity_id = self.jobid_to_entity_id(job_id)
            select_sb_id = sqlalchemy.select(
                sb_SandBoxes.SBId,
                sqlalchemy.literal(entity_id).label("EntityId"),
                sqlalchemy.literal(sb_type).label("Type"),
            ).where(
                sb_SandBoxes.SEName == se_name,
                sb_SandBoxes.SEPFN == pfn,
            )
            stmt = sqlalchemy.insert(sb_EntityMapping).from_select(
                ["SBId", "EntityId", "Type"], select_sb_id
            )
            await self.conn.execute(stmt)

            stmt = (
                sqlalchemy.update(sb_SandBoxes)
                .where(sb_SandBoxes.SEPFN == pfn)
                .values(Assigned=True)
            )
            result = await self.conn.execute(stmt)
            assert result.rowcount == 1

    async def unassign_sandboxes_to_jobs(self, jobs_ids: list[int]) -> None:
        """Delete mapping between jobs and sandboxes."""
        for job_id in jobs_ids:
            entity_id = self.jobid_to_entity_id(job_id)
            sb_sel_stmt = sqlalchemy.select(sb_SandBoxes.SBId)
            sb_sel_stmt = sb_sel_stmt.join(
                sb_EntityMapping, sb_EntityMapping.SBId == sb_SandBoxes.SBId
            )
            sb_sel_stmt = sb_sel_stmt.where(sb_EntityMapping.EntityId == entity_id)

            result = await self.conn.execute(sb_sel_stmt)
            sb_ids = [row.SBId for row in result]

            del_stmt = sqlalchemy.delete(sb_EntityMapping).where(
                sb_EntityMapping.EntityId == entity_id
            )
            await self.conn.execute(del_stmt)

            sb_entity_sel_stmt = sqlalchemy.select(sb_EntityMapping.SBId).where(
                sb_EntityMapping.SBId.in_(sb_ids)
            )
            result = await self.conn.execute(sb_entity_sel_stmt)
            remaining_sb_ids = [row.SBId for row in result]
            if not remaining_sb_ids:
                unassign_stmt = (
                    sqlalchemy.update(sb_SandBoxes)
                    .where(sb_SandBoxes.SBId.in_(sb_ids))
                    .values(Assigned=False)
                )
                await self.conn.execute(unassign_stmt)



from sqlalchemy import (
    DateTime,
    Enum,
    ForeignKey,
    Index,
    Integer,
    String,
    Text,
)
from sqlalchemy.orm import declarative_base

from ..utils import Column, EnumBackedBool, NullColumn

JobDBBase = declarative_base()


class Jobs(JobDBBase):
    __tablename__ = "Jobs"

    JobID = Column(
        "JobID",
        Integer,
        ForeignKey("JobJDLs.JobID", ondelete="CASCADE"),
        primary_key=True,
        default=0,
    )
    JobType = Column("JobType", String(32), default="user")
    JobGroup = Column("JobGroup", String(32), default="00000000")
    Site = Column("Site", String(100), default="ANY")
    JobName = Column("JobName", String(128), default="Unknown")
    Owner = Column("Owner", String(64), default="Unknown")
    OwnerGroup = Column("OwnerGroup", String(128), default="Unknown")
    VO = Column("VO", String(32))
    SubmissionTime = NullColumn("SubmissionTime", DateTime)
    RescheduleTime = NullColumn("RescheduleTime", DateTime)
    LastUpdateTime = NullColumn("LastUpdateTime", DateTime)
    StartExecTime = NullColumn("StartExecTime", DateTime)
    HeartBeatTime = NullColumn("HeartBeatTime", DateTime)
    EndExecTime = NullColumn("EndExecTime", DateTime)
    Status = Column("Status", String(32), default="Received")
    MinorStatus = Column("MinorStatus", String(128), default="Unknown")
    ApplicationStatus = Column("ApplicationStatus", String(255), default="Unknown")
    UserPriority = Column("UserPriority", Integer, default=0)
    RescheduleCounter = Column("RescheduleCounter", Integer, default=0)
    VerifiedFlag = Column("VerifiedFlag", EnumBackedBool(), default=False)
    # TODO: Should this be True/False/"Failed"? Or True/False/Null?
    AccountedFlag = Column(
        "AccountedFlag", Enum("True", "False", "Failed"), default="False"
    )

    __table_args__ = (
        Index("JobType", "JobType"),
        Index("JobGroup", "JobGroup"),
        Index("Site", "Site"),
        Index("Owner", "Owner"),
        Index("OwnerGroup", "OwnerGroup"),
        Index("Status", "Status"),
        Index("MinorStatus", "MinorStatus"),
        Index("ApplicationStatus", "ApplicationStatus"),
        Index("StatusSite", "Status", "Site"),
        Index("LastUpdateTime", "LastUpdateTime"),
    )


class JobJDLs(JobDBBase):
    __tablename__ = "JobJDLs"
    JobID = Column(Integer, autoincrement=True, primary_key=True)
    JDL = Column(Text)
    JobRequirements = Column(Text)
    OriginalJDL = Column(Text)


class InputData(JobDBBase):
    __tablename__ = "InputData"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    LFN = Column(String(255), default="", primary_key=True)
    Status = Column(String(32), default="AprioriGood")


class JobParameters(JobDBBase):
    __tablename__ = "JobParameters"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    Name = Column(String(100), primary_key=True)
    Value = Column(Text)


class OptimizerParameters(JobDBBase):
    __tablename__ = "OptimizerParameters"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    Name = Column(String(100), primary_key=True)
    Value = Column(Text)


class AtticJobParameters(JobDBBase):
    __tablename__ = "AtticJobParameters"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    Name = Column(String(100), primary_key=True)
    Value = Column(Text)
    RescheduleCycle = Column(Integer)


class HeartBeatLoggingInfo(JobDBBase):
    __tablename__ = "HeartBeatLoggingInfo"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    Name = Column(String(100), primary_key=True)
    Value = Column(Text)
    HeartBeatTime = Column(DateTime, primary_key=True)


class JobCommands(JobDBBase):
    __tablename__ = "JobCommands"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    Command = Column(String(100))
    Arguments = Column(String(100))
    Status = Column(String(64), default="Received")
    ReceptionTime = Column(DateTime, primary_key=True)
    ExecutionTime = NullColumn(DateTime)



from __future__ import annotations

import time
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from sqlalchemy import delete, func, insert, select

if TYPE_CHECKING:
    pass

from diracx.core.exceptions import JobNotFound
from diracx.core.models import (
    JobStatus,
    JobStatusReturn,
)

from ..utils import BaseSQLDB
from .schema import (
    JobLoggingDBBase,
    LoggingInfo,
)

MAGIC_EPOC_NUMBER = 1270000000


class JobLoggingDB(BaseSQLDB):
    """Frontend for the JobLoggingDB. Provides the ability to store changes with timestamps."""

    metadata = JobLoggingDBBase.metadata

    async def insert_record(
        self,
        job_id: int,
        status: JobStatus,
        minor_status: str,
        application_status: str,
        date: datetime,
        source: str,
    ):
        """Add a new entry to the JobLoggingDB table. One, two or all the three status
        components (status, minorStatus, applicationStatus) can be specified.
        Optionally the time stamp of the status can
        be provided in a form of a string in a format '%Y-%m-%d %H:%M:%S' or
        as datetime.datetime object. If the time stamp is not provided the current
        UTC time is used.
        """
        # First, fetch the maximum SeqNum for the given job_id
        seqnum_stmt = select(func.coalesce(func.max(LoggingInfo.SeqNum) + 1, 1)).where(
            LoggingInfo.JobID == job_id
        )
        seqnum = await self.conn.scalar(seqnum_stmt)

        epoc = (
            time.mktime(date.timetuple())
            + date.microsecond / 1000000.0
            - MAGIC_EPOC_NUMBER
        )

        stmt = insert(LoggingInfo).values(
            JobID=int(job_id),
            SeqNum=seqnum,
            Status=status,
            MinorStatus=minor_status,
            ApplicationStatus=application_status[:255],
            StatusTime=date,
            StatusTimeOrder=epoc,
            Source=source[:32],
        )
        await self.conn.execute(stmt)

    async def get_records(self, job_id: int) -> list[JobStatusReturn]:
        """Returns a Status,MinorStatus,ApplicationStatus,StatusTime,Source tuple
        for each record found for job specified by its jobID in historical order.
        """
        stmt = (
            select(
                LoggingInfo.Status,
                LoggingInfo.MinorStatus,
                LoggingInfo.ApplicationStatus,
                LoggingInfo.StatusTime,
                LoggingInfo.Source,
            )
            .where(LoggingInfo.JobID == int(job_id))
            .order_by(LoggingInfo.StatusTimeOrder, LoggingInfo.StatusTime)
        )
        rows = await self.conn.execute(stmt)

        values = []
        for (
            status,
            minor_status,
            application_status,
            status_time,
            status_source,
        ) in rows:
            values.append(
                [
                    status,
                    minor_status,
                    application_status,
                    status_time.replace(tzinfo=timezone.utc),
                    status_source,
                ]
            )

        # If no value has been set for the application status in the first place,
        # We put this status to unknown
        res = []
        if values:
            if values[0][2] == "idem":
                values[0][2] = "Unknown"

            # We replace "idem" values by the value previously stated
            for i in range(1, len(values)):
                for j in range(3):
                    if values[i][j] == "idem":
                        values[i][j] = values[i - 1][j]

            # And we replace arrays with tuples
            for (
                status,
                minor_status,
                application_status,
                status_time,
                status_source,
            ) in values:
                res.append(
                    JobStatusReturn(
                        Status=status,
                        MinorStatus=minor_status,
                        ApplicationStatus=application_status,
                        StatusTime=status_time,
                        Source=status_source,
                    )
                )

        return res

    async def delete_records(self, job_ids: list[int]):
        """Delete logging records for given jobs."""
        stmt = delete(LoggingInfo).where(LoggingInfo.JobID.in_(job_ids))
        await self.conn.execute(stmt)

    async def get_wms_time_stamps(self, job_id):
        """Get TimeStamps for job MajorState transitions
        return a {State:timestamp} dictionary.
        """
        result = {}
        stmt = select(
            LoggingInfo.Status,
            LoggingInfo.StatusTimeOrder,
        ).where(LoggingInfo.JobID == job_id)
        rows = await self.conn.execute(stmt)
        if not rows.rowcount:
            raise JobNotFound(job_id) from None

        for event, etime in rows:
            result[event] = str(etime + MAGIC_EPOC_NUMBER)

        return result
